@@ -4,8 +4,10 @@ use crate::{
     body::{box_body, BoxBody},
     extract::FromRequest,
     response::IntoResponse,
-    routing::{EmptyRouter, MethodFilter, RouteFuture},
+    routing::{EmptyRouter, MethodFilter},
     service::HandleError,
+    util::Either,
+    BoxError,
 };
 use async_trait::async_trait;
 use bytes::Bytes;
@@ -17,9 +19,14 @@ use std::{
     marker::PhantomData,
     task::{Context, Poll},
 };
-use tower::{BoxError, Layer, Service, ServiceExt};
+use tower::ServiceExt;
+use tower_layer::Layer;
+use tower_service::Service;
 
 pub mod future;
+mod into_service;
+
+pub use self::into_service::IntoService;
 
 /// Route requests to the given handler regardless of the HTTP method of the
 /// request.
@@ -27,41 +34,44 @@ pub mod future;
 /// # Example
 ///
 /// ```rust
-/// use axum::prelude::*;
+/// use axum::{
+///     handler::any,
+///     Router,
+/// };
 ///
 /// async fn handler() {}
 ///
 /// // All requests to `/` will go to `handler` regardless of the HTTP method.
-/// let app = route("/", any(handler));
+/// let app = Router::new().route("/", any(handler));
 /// # async {
-/// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+/// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 /// # };
 /// ```
-pub fn any<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn any<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Any, handler)
+    on(MethodFilter::all(), handler)
 }
 
 /// Route `CONNECT` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn connect<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn connect<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Connect, handler)
+    on(MethodFilter::CONNECT, handler)
 }
 
 /// Route `DELETE` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn delete<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn delete<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Delete, handler)
+    on(MethodFilter::DELETE, handler)
 }
 
 /// Route `GET` requests to the given handler.
@@ -69,81 +79,88 @@ where
 /// # Example
 ///
 /// ```rust
-/// use axum::prelude::*;
+/// use axum::{
+///     handler::get,
+///     Router,
+/// };
 ///
 /// async fn handler() {}
 ///
 /// // Requests to `GET /` will go to `handler`.
-/// let app = route("/", get(handler));
+/// let app = Router::new().route("/", get(handler));
 /// # async {
-/// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+/// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 /// # };
 /// ```
-pub fn get<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+///
+/// Note that `get` routes will also be called for `HEAD` requests but will have
+/// the response body removed. Make sure to add explicit `HEAD` routes
+/// afterwards.
+pub fn get<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Get, handler)
+    on(MethodFilter::GET | MethodFilter::HEAD, handler)
 }
 
 /// Route `HEAD` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn head<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn head<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Head, handler)
+    on(MethodFilter::HEAD, handler)
 }
 
 /// Route `OPTIONS` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn options<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn options<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Options, handler)
+    on(MethodFilter::OPTIONS, handler)
 }
 
 /// Route `PATCH` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn patch<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn patch<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Patch, handler)
+    on(MethodFilter::PATCH, handler)
 }
 
 /// Route `POST` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn post<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn post<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Post, handler)
+    on(MethodFilter::POST, handler)
 }
 
 /// Route `PUT` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn put<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn put<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Put, handler)
+    on(MethodFilter::PUT, handler)
 }
 
 /// Route `TRACE` requests to the given handler.
 ///
 /// See [`get`] for an example.
-pub fn trace<H, B, T>(handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn trace<H, B, T>(handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
-    on(MethodFilter::Trace, handler)
+    on(MethodFilter::TRACE, handler)
 }
 
 /// Route requests with the given method to the handler.
@@ -151,24 +168,29 @@ where
 /// # Example
 ///
 /// ```rust
-/// use axum::{handler::on, routing::MethodFilter, prelude::*};
+/// use axum::{
+///     handler::on,
+///     Router,
+///     routing::MethodFilter,
+/// };
 ///
 /// async fn handler() {}
 ///
 /// // Requests to `POST /` will go to `handler`.
-/// let app = route("/", on(MethodFilter::Post, handler));
+/// let app = Router::new().route("/", on(MethodFilter::POST, handler));
 /// # async {
-/// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+/// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 /// # };
 /// ```
-pub fn on<H, B, T>(method: MethodFilter, handler: H) -> OnMethod<IntoService<H, B, T>, EmptyRouter>
+pub fn on<H, B, T>(method: MethodFilter, handler: H) -> OnMethod<H, B, T, EmptyRouter>
 where
     H: Handler<B, T>,
 {
     OnMethod {
         method,
-        svc: handler.into_service(),
+        handler,
         fallback: EmptyRouter::method_not_allowed(),
+        _marker: PhantomData,
     }
 }
 
@@ -187,7 +209,7 @@ pub(crate) mod sealed {
 ///
 /// See the [module docs](crate::handler) for more details.
 #[async_trait]
-pub trait Handler<B, In>: Sized {
+pub trait Handler<B, T>: Clone + Send + Sized + 'static {
     // This seals the trait. We cannot use the regular "sealed super trait"
     // approach due to coherence.
     #[doc(hidden)]
@@ -204,7 +226,7 @@ pub trait Handler<B, In>: Sized {
     /// This can be used to add additional processing to a request for a single
     /// handler.
     ///
-    /// Note this differes from [`routing::Layered`](crate::routing::Layered)
+    /// Note this differs from [`routing::Layered`](crate::routing::Layered)
     /// which adds a middleware to a group of routes.
     ///
     /// # Example
@@ -213,29 +235,55 @@ pub trait Handler<B, In>: Sized {
     /// can be done like so:
     ///
     /// ```rust
-    /// use axum::prelude::*;
+    /// use axum::{
+    ///     handler::{get, Handler},
+    ///     Router,
+    /// };
     /// use tower::limit::{ConcurrencyLimitLayer, ConcurrencyLimit};
     ///
     /// async fn handler() { /* ... */ }
     ///
     /// let layered_handler = handler.layer(ConcurrencyLimitLayer::new(64));
-    /// let app = route("/", get(layered_handler));
+    /// let app = Router::new().route("/", get(layered_handler));
     /// # async {
-    /// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+    /// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
     /// # };
     /// ```
     ///
     /// When adding middleware that might fail its recommended to handle those
     /// errors. See [`Layered::handle_error`] for more details.
-    fn layer<L>(self, layer: L) -> Layered<L::Service, In>
+    fn layer<L>(self, layer: L) -> Layered<L::Service, T>
     where
-        L: Layer<IntoService<Self, B, In>>,
+        L: Layer<OnMethod<Self, B, T, EmptyRouter>>,
     {
-        Layered::new(layer.layer(IntoService::new(self)))
+        Layered::new(layer.layer(any(self)))
     }
 
     /// Convert the handler into a [`Service`].
-    fn into_service(self) -> IntoService<Self, B, In> {
+    ///
+    /// This allows you to serve a single handler if you don't need any routing:
+    ///
+    /// ```rust
+    /// use axum::{
+    ///     Server, handler::Handler, http::{Uri, Method}, response::IntoResponse,
+    /// };
+    /// use tower::make::Shared;
+    /// use std::net::SocketAddr;
+    ///
+    /// async fn handler(method: Method, uri: Uri, body: String) -> impl IntoResponse {
+    ///     format!("received `{} {}` with body `{:?}`", method, uri, body)
+    /// }
+    ///
+    /// let service = handler.into_service();
+    ///
+    /// # async {
+    /// Server::bind(&SocketAddr::from(([127, 0, 0, 1], 3000)))
+    ///     .serve(Shared::new(service))
+    ///     .await?;
+    /// # Ok::<_, hyper::Error>(())
+    /// # };
+    /// ```
+    fn into_service(self) -> IntoService<Self, B, T> {
         IntoService::new(self)
     }
 }
@@ -243,7 +291,7 @@ pub trait Handler<B, In>: Sized {
 #[async_trait]
 impl<F, Fut, Res, B> Handler<B, ()> for F
 where
-    F: FnOnce() -> Fut + Send + Sync,
+    F: FnOnce() -> Fut + Clone + Send + Sync + 'static,
     Fut: Future<Output = Res> + Send,
     Res: IntoResponse,
     B: Send + 'static,
@@ -264,12 +312,13 @@ macro_rules! impl_handler {
         #[allow(non_snake_case)]
         impl<F, Fut, B, Res, $head, $($tail,)*> Handler<B, ($head, $($tail,)*)> for F
         where
-            F: FnOnce($head, $($tail,)*) -> Fut + Send + Sync,
+            F: FnOnce($head, $($tail,)*) -> Fut + Clone + Send + Sync + 'static,
             Fut: Future<Output = Res> + Send,
             B: Send + 'static,
             Res: IntoResponse,
+            B: Send + 'static,
             $head: FromRequest<B> + Send,
-            $( $tail: FromRequest<B> + Send, )*
+            $( $tail: FromRequest<B> + Send,)*
         {
             type Sealed = sealed::Hidden;
 
@@ -278,13 +327,13 @@ macro_rules! impl_handler {
 
                 let $head = match $head::from_request(&mut req).await {
                     Ok(value) => value,
-                    Err(rejection) => return rejection.into_response().map(crate::body::box_body),
+                    Err(rejection) => return rejection.into_response().map(box_body),
                 };
 
                 $(
                     let $tail = match $tail::from_request(&mut req).await {
                         Ok(value) => value,
-                        Err(rejection) => return rejection.into_response().map(crate::body::box_body),
+                        Err(rejection) => return rejection.into_response().map(box_body),
                     };
                 )*
 
@@ -329,9 +378,10 @@ where
 #[async_trait]
 impl<S, T, ReqBody, ResBody> Handler<ReqBody, T> for Layered<S, T>
 where
-    S: Service<Request<ReqBody>, Response = Response<ResBody>> + Send,
+    S: Service<Request<ReqBody>, Response = Response<ResBody>> + Clone + Send + 'static,
     S::Error: IntoResponse,
     S::Future: Send,
+    T: 'static,
     ReqBody: Send + 'static,
     ResBody: http_body::Body<Data = Bytes> + Send + Sync + 'static,
     ResBody::Error: Into<BoxError> + Send + Sync + 'static,
@@ -365,9 +415,9 @@ impl<S, T> Layered<S, T> {
     /// This is used to convert errors to responses rather than simply
     /// terminating the connection.
     ///
-    /// It works similarly to [`routing::Layered::handle_error`]. See that for more details.
+    /// It works similarly to [`routing::Router::handle_error`]. See that for more details.
     ///
-    /// [`routing::Layered::handle_error`]: crate::routing::Layered::handle_error
+    /// [`routing::Router::handle_error`]: crate::routing::Router::handle_error
     pub fn handle_error<F, ReqBody, ResBody, Res, E>(
         self,
         f: F,
@@ -382,111 +432,81 @@ impl<S, T> Layered<S, T> {
     }
 }
 
-/// An adapter that makes a [`Handler`] into a [`Service`].
-///
-/// Created with [`Handler::into_service`].
-pub struct IntoService<H, B, T> {
-    handler: H,
-    _marker: PhantomData<fn() -> (B, T)>,
+/// A handler [`Service`] that accepts requests based on a [`MethodFilter`] and
+/// allows chaining additional handlers.
+pub struct OnMethod<H, B, T, F> {
+    pub(crate) method: MethodFilter,
+    pub(crate) handler: H,
+    pub(crate) fallback: F,
+    pub(crate) _marker: PhantomData<fn() -> (B, T)>,
 }
 
-impl<H, B, T> IntoService<H, B, T> {
-    fn new(handler: H) -> Self {
-        Self {
-            handler,
-            _marker: PhantomData,
-        }
-    }
-}
-
-impl<H, B, T> fmt::Debug for IntoService<H, B, T>
+impl<H, B, T, F> fmt::Debug for OnMethod<H, B, T, F>
 where
-    H: fmt::Debug,
+    T: fmt::Debug,
+    F: fmt::Debug,
 {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("IntoService")
-            .field("handler", &self.handler)
+        f.debug_struct("OnMethod")
+            .field("method", &self.method)
+            .field("handler", &format_args!("{}", std::any::type_name::<H>()))
+            .field("fallback", &self.fallback)
             .finish()
     }
 }
 
-impl<H, B, T> Clone for IntoService<H, B, T>
+impl<H, B, T, F> Clone for OnMethod<H, B, T, F>
 where
     H: Clone,
+    F: Clone,
 {
     fn clone(&self) -> Self {
         Self {
+            method: self.method,
             handler: self.handler.clone(),
+            fallback: self.fallback.clone(),
             _marker: PhantomData,
         }
     }
 }
 
-impl<H, T, B> Service<Request<B>> for IntoService<H, B, T>
+impl<H, B, T, F> Copy for OnMethod<H, B, T, F>
 where
-    H: Handler<B, T> + Clone + Send + 'static,
-    B: Send + 'static,
+    H: Copy,
+    F: Copy,
 {
-    type Response = Response<BoxBody>;
-    type Error = Infallible;
-    type Future = future::IntoServiceFuture;
-
-    fn poll_ready(&mut self, _cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
-        // `IntoService` can only be constructed from async functions which are always ready, or from
-        // `Layered` which bufferes in `<Layered as Handler>::call` and is therefore also always
-        // ready.
-        Poll::Ready(Ok(()))
-    }
-
-    fn call(&mut self, req: Request<B>) -> Self::Future {
-        let handler = self.handler.clone();
-        let future = Box::pin(async move {
-            let res = Handler::call(handler, req).await;
-            Ok(res)
-        });
-        future::IntoServiceFuture(future)
-    }
 }
 
-/// A handler [`Service`] that accepts requests based on a [`MethodFilter`] and
-/// allows chaining additional handlers.
-#[derive(Debug, Clone, Copy)]
-pub struct OnMethod<S, F> {
-    pub(crate) method: MethodFilter,
-    pub(crate) svc: S,
-    pub(crate) fallback: F,
-}
-
-impl<S, F> OnMethod<S, F> {
+impl<H, B, T, F> OnMethod<H, B, T, F> {
     /// Chain an additional handler that will accept all requests regardless of
     /// its HTTP method.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn any<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn any<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Any, handler)
+        self.on(MethodFilter::all(), handler)
     }
 
     /// Chain an additional handler that will only accept `CONNECT` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn connect<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn connect<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Connect, handler)
+        self.on(MethodFilter::CONNECT, handler)
     }
 
     /// Chain an additional handler that will only accept `DELETE` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn delete<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn delete<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Delete, handler)
+        self.on(MethodFilter::DELETE, handler)
     }
 
     /// Chain an additional handler that will only accept `GET` requests.
@@ -494,7 +514,7 @@ impl<S, F> OnMethod<S, F> {
     /// # Example
     ///
     /// ```rust
-    /// use axum::prelude::*;
+    /// use axum::{handler::post, Router};
     ///
     /// async fn handler() {}
     ///
@@ -502,76 +522,80 @@ impl<S, F> OnMethod<S, F> {
     ///
     /// // Requests to `GET /` will go to `handler` and `POST /` will go to
     /// // `other_handler`.
-    /// let app = route("/", post(handler).get(other_handler));
+    /// let app = Router::new().route("/", post(handler).get(other_handler));
     /// # async {
-    /// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+    /// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
     /// # };
     /// ```
-    pub fn get<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    ///
+    /// Note that `get` routes will also be called for `HEAD` requests but will have
+    /// the response body removed. Make sure to add explicit `HEAD` routes
+    /// afterwards.
+    pub fn get<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Get, handler)
+        self.on(MethodFilter::GET | MethodFilter::HEAD, handler)
     }
 
     /// Chain an additional handler that will only accept `HEAD` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn head<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn head<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Head, handler)
+        self.on(MethodFilter::HEAD, handler)
     }
 
     /// Chain an additional handler that will only accept `OPTIONS` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn options<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn options<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Options, handler)
+        self.on(MethodFilter::OPTIONS, handler)
     }
 
     /// Chain an additional handler that will only accept `PATCH` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn patch<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn patch<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Patch, handler)
+        self.on(MethodFilter::PATCH, handler)
     }
 
     /// Chain an additional handler that will only accept `POST` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn post<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn post<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Post, handler)
+        self.on(MethodFilter::POST, handler)
     }
 
     /// Chain an additional handler that will only accept `PUT` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn put<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn put<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Put, handler)
+        self.on(MethodFilter::PUT, handler)
     }
 
     /// Chain an additional handler that will only accept `TRACE` requests.
     ///
     /// See [`OnMethod::get`] for an example.
-    pub fn trace<H, B, T>(self, handler: H) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn trace<H2, T2>(self, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
-        self.on(MethodFilter::Trace, handler)
+        self.on(MethodFilter::TRACE, handler)
     }
 
     /// Chain an additional handler that will accept requests matching the given
@@ -580,7 +604,11 @@ impl<S, F> OnMethod<S, F> {
     /// # Example
     ///
     /// ```rust
-    /// use axum::{routing::MethodFilter, prelude::*};
+    /// use axum::{
+    ///     handler::get,
+    ///     Router,
+    ///     routing::MethodFilter
+    /// };
     ///
     /// async fn handler() {}
     ///
@@ -588,47 +616,52 @@ impl<S, F> OnMethod<S, F> {
     ///
     /// // Requests to `GET /` will go to `handler` and `DELETE /` will go to
     /// // `other_handler`
-    /// let app = route("/", get(handler).on(MethodFilter::Delete, other_handler));
+    /// let app = Router::new().route("/", get(handler).on(MethodFilter::DELETE, other_handler));
     /// # async {
-    /// # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+    /// # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
     /// # };
     /// ```
-    pub fn on<H, B, T>(
-        self,
-        method: MethodFilter,
-        handler: H,
-    ) -> OnMethod<IntoService<H, B, T>, Self>
+    pub fn on<H2, T2>(self, method: MethodFilter, handler: H2) -> OnMethod<H2, B, T2, Self>
     where
-        H: Handler<B, T>,
+        H2: Handler<B, T2>,
     {
         OnMethod {
             method,
-            svc: handler.into_service(),
+            handler,
             fallback: self,
+            _marker: PhantomData,
         }
     }
 }
 
-impl<S, F, B> Service<Request<B>> for OnMethod<S, F>
+impl<H, B, T, F> Service<Request<B>> for OnMethod<H, B, T, F>
 where
-    S: Service<Request<B>, Response = Response<BoxBody>, Error = Infallible> + Clone,
+    H: Handler<B, T>,
     F: Service<Request<B>, Response = Response<BoxBody>, Error = Infallible> + Clone,
+    B: Send + 'static,
 {
     type Response = Response<BoxBody>;
     type Error = Infallible;
-    type Future = RouteFuture<S, F, B>;
+    type Future = future::OnMethodFuture<F, B>;
 
     fn poll_ready(&mut self, _cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
         Poll::Ready(Ok(()))
     }
 
     fn call(&mut self, req: Request<B>) -> Self::Future {
-        if self.method.matches(req.method()) {
-            let fut = self.svc.clone().oneshot(req);
-            RouteFuture::a(fut)
+        let req_method = req.method().clone();
+
+        let fut = if self.method.matches(req.method()) {
+            let fut = Handler::call(self.handler.clone(), req);
+            Either::A { inner: fut }
         } else {
             let fut = self.fallback.clone().oneshot(req);
-            RouteFuture::b(fut)
+            Either::B { inner: fut }
+        };
+
+        future::OnMethodFuture {
+            inner: fut,
+            req_method,
         }
     }
 }
