@@ -8,15 +8,19 @@
 //! - [Routing](#routing)
 //!     - [Precedence](#precedence)
 //!     - [Matching multiple methods](#matching-multiple-methods)
+//!     - [Routing to any `Service`](#routing-to-any-service)
+//!         - [Routing to fallible services](#routing-to-fallible-services)
+//!     - [Nesting routes](#nesting-routes)
 //! - [Extractors](#extractors)
 //! - [Building responses](#building-responses)
 //! - [Applying middleware](#applying-middleware)
 //!     - [To individual handlers](#to-individual-handlers)
 //!     - [To groups of routes](#to-groups-of-routes)
 //!     - [Error handling](#error-handling)
+//!     - [Applying multiple middleware](#applying-multiple-middleware)
+//!     - [Commonly used middleware](#commonly-used-middleware)
+//!     - [Writing your own middleware](#writing-your-own-middleware)
 //! - [Sharing state with handlers](#sharing-state-with-handlers)
-//! - [Routing to any `Service`](#routing-to-any-service)
-//! - [Nesting applications](#nesting-applications)
 //! - [Required dependencies](#required-dependencies)
 //! - [Examples](#examples)
 //! - [Feature flags](#feature-flags)
@@ -46,18 +50,18 @@
 //! The "Hello, World!" of axum is:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
-//! use hyper::Server;
-//! use std::net::SocketAddr;
-//! use tower::make::Shared;
+//! use axum::{
+//!     handler::get,
+//!     Router,
+//! };
 //!
 //! #[tokio::main]
 //! async fn main() {
 //!     // build our application with a single route
-//!     let app = route("/", get(|| async { "Hello, World!" }));
+//!     let app = Router::new().route("/", get(|| async { "Hello, World!" }));
 //!
 //!     // run it with hyper on localhost:3000
-//!     hyper::Server::bind(&"0.0.0.0:3000".parse().unwrap())
+//!     axum::Server::bind(&"0.0.0.0:3000".parse().unwrap())
 //!         .serve(app.into_make_service())
 //!         .await
 //!         .unwrap();
@@ -76,7 +80,6 @@
 //! Some examples of handlers:
 //!
 //! ```rust
-//! use axum::prelude::*;
 //! use bytes::Bytes;
 //! use http::StatusCode;
 //!
@@ -104,9 +107,13 @@
 //! Routing between handlers looks like this:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     handler::get,
+//!     Router,
+//! };
 //!
-//! let app = route("/", get(get_slash).post(post_slash))
+//! let app = Router::new()
+//!     .route("/", get(get_slash).post(post_slash))
 //!     .route("/foo", get(get_foo));
 //!
 //! async fn get_slash() {
@@ -121,21 +128,28 @@
 //!     // `GET /foo` called
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
 //! Routes can also be dynamic like `/users/:id`. See [extractors](#extractors)
 //! for more details.
 //!
+//! You can also define routes separately and merge them with [`Router::or`].
+//!
 //! ## Precedence
 //!
 //! Note that routes are matched _bottom to top_ so routes that should have
 //! higher precedence should be added _after_ routes with lower precedence:
 //!
 //! ```rust
-//! use axum::{prelude::*, body::BoxBody};
-//! use tower::{Service, ServiceExt, BoxError};
+//! use axum::{
+//!     body::{Body, BoxBody},
+//!     handler::get,
+//!     http::Request,
+//!     Router,
+//! };
+//! use tower::{Service, ServiceExt};
 //! use http::{Method, Response, StatusCode};
 //! use std::convert::Infallible;
 //!
@@ -143,7 +157,8 @@
 //! # async fn main() {
 //! // `/foo` also matches `/:key` so adding the routes in this order means `/foo`
 //! // will be inaccessible.
-//! let mut app = route("/foo", get(|| async { "/foo called" }))
+//! let mut app = Router::new()
+//!     .route("/foo", get(|| async { "/foo called" }))
 //!     .route("/:key", get(|| async { "/:key called" }));
 //!
 //! // Even though we use `/foo` as the request URI, `/:key` takes precedence
@@ -154,7 +169,8 @@
 //!
 //! // We have to add `/foo` after `/:key` since routes are matched bottom to
 //! // top.
-//! let mut new_app = route("/:key", get(|| async { "/:key called" }))
+//! let mut new_app = Router::new()
+//!     .route("/:key", get(|| async { "/:key called" }))
 //!     .route("/foo", get(|| async { "/foo called" }));
 //!
 //! // Now it works
@@ -191,27 +207,169 @@
 //! # }
 //! ```
 //!
-//! ## Matching multiple methods
+//! ## Routing to any [`Service`]
 //!
-//! If you want a path to accept multiple HTTP methods you must add them all at
-//! once:
+//! axum also supports routing to general [`Service`]s:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     body::Body,
+//!     http::Request,
+//!     Router,
+//!     service
+//! };
+//! use tower_http::services::ServeFile;
+//! use http::Response;
+//! use std::convert::Infallible;
+//! use tower::service_fn;
 //!
-//! // `GET /` and `POST /` are both accepted
-//! let app = route("/", get(handler).post(handler));
-//!
-//! // This will _not_ work. Only `POST /` will be accessible.
-//! let wont_work = route("/", get(handler)).route("/", post(handler));
-//!
-//! async fn handler() {}
+//! let app = Router::new()
+//!     .route(
+//!         // Any request to `/` goes to a service
+//!         "/",
+//!         // Services who's response body is not `axum::body::BoxBody`
+//!         // can be wrapped in `axum::service::any` (or one of the other routing filters)
+//!         // to have the response body mapped
+//!         service::any(service_fn(|_: Request<Body>| async {
+//!             let res = Response::new(Body::from("Hi from `GET /`"));
+//!             Ok(res)
+//!         }))
+//!     )
+//!     .route(
+//!         "/foo",
+//!         // This service's response body is `axum::body::BoxBody` so
+//!         // it can be routed to directly.
+//!         service_fn(|req: Request<Body>| async move {
+//!             let body = Body::from(format!("Hi from `{} /foo`", req.method()));
+//!             let body = axum::body::box_body(body);
+//!             let res = Response::new(body);
+//!             Ok(res)
+//!         })
+//!     )
+//!     .route(
+//!         // GET `/static/Cargo.toml` goes to a service from tower-http
+//!         "/static/Cargo.toml",
+//!         service::get(ServeFile::new("Cargo.toml"))
+//!     );
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(wont_work.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
+//! Routing to arbitrary services in this way has complications for backpressure
+//! ([`Service::poll_ready`]). See the [`service`] module for more details.
+//!
+//! ### Routing to fallible services
+//!
+//! Note that routing to general services has a small gotcha when it comes to
+//! errors. axum currently does not support mixing routes to fallible services
+//! with infallible handlers. For example this does _not_ compile:
+//!
+//! ```compile_fail
+//! use axum::{
+//!     Router,
+//!     service,
+//!     handler::get,
+//!     http::{Request, Response},
+//!     body::Body,
+//! };
+//! use std::io;
+//! use tower::service_fn;
+//!
+//! let app = Router::new()
+//!     // this route cannot fail
+//!     .route("/foo", get(|| async {}))
+//!     // this route can fail with io::Error
+//!     .route(
+//!         "/",
+//!         service::get(service_fn(|_req: Request<Body>| async {
+//!             let contents = tokio::fs::read_to_string("some_file").await?;
+//!             Ok::<_, io::Error>(Response::new(Body::from(contents)))
+//!         })),
+//!     );
+//! # async {
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # };
+//! ```
+//!
+//! The solution is to use [`handle_error`] and handle the error from the
+//! service:
+//!
+//! ```
+//! use axum::{
+//!     Router,
+//!     service,
+//!     handler::get,
+//!     http::{Request, Response},
+//!     response::IntoResponse,
+//!     body::Body,
+//! };
+//! use std::{io, convert::Infallible};
+//! use tower::service_fn;
+//!
+//! let app = Router::new()
+//!     // this route cannot fail
+//!     .route("/foo", get(|| async {}))
+//!     // this route can fail with io::Error
+//!     .route(
+//!         "/",
+//!         service::get(service_fn(|_req: Request<Body>| async {
+//!             let contents = tokio::fs::read_to_string("some_file").await?;
+//!             Ok::<_, io::Error>(Response::new(Body::from(contents)))
+//!         }))
+//!         .handle_error(handle_io_error),
+//!     );
+//!
+//! fn handle_io_error(error: io::Error) -> Result<impl IntoResponse, Infallible> {
+//!     # Ok(())
+//!     // ...
+//! }
+//! # async {
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # };
+//! ```
+//!
+//! In this particular case you can also handle the error directly in
+//! `service_fn` but that is not possible, if you're routing to a service which
+//! you don't control.
+//!
+//! See ["Error handling"](#error-handling) for more details on [`handle_error`]
+//! and error handling in general.
+//!
+//! ## Nesting routes
+//!
+//! Routes can be nested by calling [`Router::nest`](routing::Router::nest):
+//!
+//! ```rust,no_run
+//! use axum::{
+//!     body::{Body, BoxBody},
+//!     http::Request,
+//!     handler::get,
+//!     Router,
+//!     routing::BoxRoute
+//! };
+//! use tower_http::services::ServeFile;
+//! use http::Response;
+//!
+//! fn api_routes() -> Router<BoxRoute> {
+//!     Router::new()
+//!         .route("/users", get(|_: Request<Body>| async { /* ... */ }))
+//!         .boxed()
+//! }
+//!
+//! let app = Router::new()
+//!     .route("/", get(|_: Request<Body>| async { /* ... */ }))
+//!     .nest("/api", api_routes());
+//! # async {
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # };
+//! ```
+//!
+//! Note that nested routes will not see the orignal request URI but instead
+//! have the matched prefix stripped. This is necessary for services like static
+//! file serving to work. Use [`OriginalUri`] if you need the original request
+//! URI.
+//!
 //! # Extractors
 //!
 //! An extractor is a type that implements [`FromRequest`]. Extractors is how
@@ -221,10 +379,14 @@
 //! body and deserializes it as JSON into some target type:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     extract,
+//!     handler::post,
+//!     Router,
+//! };
 //! use serde::Deserialize;
 //!
-//! let app = route("/users", post(create_user));
+//! let app = Router::new().route("/users", post(create_user));
 //!
 //! #[derive(Deserialize)]
 //! struct CreateUser {
@@ -238,41 +400,44 @@
 //!     // ...
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
-//! [`extract::UrlParams`] can be used to extract params from a dynamic URL. It
-//! is compatible with any type that implements [`std::str::FromStr`], such as
+//! [`extract::Path`] can be used to extract params from a dynamic URL. It
+//! is compatible with any type that implements [`serde::Deserialize`], such as
 //! [`Uuid`]:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     extract,
+//!     handler::post,
+//!     Router,
+//! };
 //! use uuid::Uuid;
 //!
-//! let app = route("/users/:id", post(create_user));
-//!
-//! async fn create_user(params: extract::UrlParams<(Uuid,)>) {
-//!     let user_id: Uuid = (params.0).0;
+//! let app = Router::new().route("/users/:id", post(create_user));
 //!
+//! async fn create_user(extract::Path(user_id): extract::Path<Uuid>) {
 //!     // ...
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
-//! There is also [`UrlParamsMap`](extract::UrlParamsMap) which provide a map
-//! like API for extracting URL params.
-//!
 //! You can also apply multiple extractors:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     extract,
+//!     handler::get,
+//!     Router,
+//! };
 //! use uuid::Uuid;
 //! use serde::Deserialize;
 //!
-//! let app = route("/users/:id/things", get(get_user_things));
+//! let app = Router::new().route("/users/:id/things", get(get_user_things));
 //!
 //! #[derive(Deserialize)]
 //! struct Pagination {
@@ -287,31 +452,35 @@
 //! }
 //!
 //! async fn get_user_things(
-//!     params: extract::UrlParams<(Uuid,)>,
+//!     extract::Path(user_id): extract::Path<Uuid>,
 //!     pagination: Option<extract::Query<Pagination>>,
 //! ) {
-//!     let user_id: Uuid = (params.0).0;
 //!     let pagination: Pagination = pagination.unwrap_or_default().0;
 //!
 //!     // ...
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
 //! Additionally `Request<Body>` is itself an extractor:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     body::Body,
+//!     handler::post,
+//!     http::Request,
+//!     Router,
+//! };
 //!
-//! let app = route("/users/:id", post(handler));
+//! let app = Router::new().route("/users/:id", post(handler));
 //!
 //! async fn handler(req: Request<Body>) {
 //!     // ...
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
@@ -329,7 +498,13 @@
 //! returned from a handler:
 //!
 //! ```rust,no_run
-//! use axum::{body::Body, response::{Html, Json}, prelude::*};
+//! use axum::{
+//!     body::Body,
+//!     handler::{get, Handler},
+//!     http::{Request, header::{HeaderMap, HeaderName, HeaderValue}},
+//!     response::{IntoResponse, Html, Json, Headers},
+//!     Router,
+//! };
 //! use http::{StatusCode, Response, Uri};
 //! use serde_json::{Value, json};
 //!
@@ -362,6 +537,33 @@
 //!     (StatusCode::INTERNAL_SERVER_ERROR, "Something went wrong")
 //! }
 //!
+//! // A tuple of `HeaderMap` and something that implements `IntoResponse` can
+//! // be used to override the headers
+//! async fn with_headers() -> (HeaderMap, &'static str) {
+//!     let mut headers = HeaderMap::new();
+//!     headers.insert(
+//!         HeaderName::from_static("x-foo"),
+//!         HeaderValue::from_static("foo"),
+//!     );
+//!     (headers, "foo")
+//! }
+//!
+//! // You can also override both status and headers at the same time
+//! async fn with_headers_and_status() -> (StatusCode, HeaderMap, &'static str) {
+//!     let mut headers = HeaderMap::new();
+//!     headers.insert(
+//!         HeaderName::from_static("x-foo"),
+//!         HeaderValue::from_static("foo"),
+//!     );
+//!     (StatusCode::INTERNAL_SERVER_ERROR, headers, "foo")
+//! }
+//!
+//! // `Headers` makes building the header map easier and `impl Trait` is easier
+//! // so you don't have to write the whole type
+//! async fn with_easy_headers() -> impl IntoResponse {
+//!     Headers(vec![("x-foo", "foo")])
+//! }
+//!
 //! // `Html` gives a content-type of `text/html`
 //! async fn html() -> Html<&'static str> {
 //!     Html("<h1>Hello, World!</h1>")
@@ -384,42 +586,53 @@
 //!     Response::builder().body(Body::empty()).unwrap()
 //! }
 //!
-//! let app = route("/plain_text", get(plain_text))
+//! let app = Router::new()
+//!     .route("/plain_text", get(plain_text))
 //!     .route("/plain_text_string", get(plain_text_string))
 //!     .route("/bytes", get(bytes))
 //!     .route("/empty", get(empty))
 //!     .route("/empty_with_status", get(empty_with_status))
 //!     .route("/with_status", get(with_status))
+//!     .route("/with_headers", get(with_headers))
+//!     .route("/with_headers_and_status", get(with_headers_and_status))
+//!     .route("/with_easy_headers", get(with_easy_headers))
 //!     .route("/html", get(html))
 //!     .route("/json", get(json))
 //!     .route("/result", get(result))
 //!     .route("/response", get(response));
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
 //! # Applying middleware
 //!
 //! axum is designed to take full advantage of the tower and tower-http
-//! ecosystem of middleware:
+//! ecosystem of middleware.
+//!
+//! If you're new to tower we recommend you read its [guides][tower-guides] for
+//! a general introduction to tower and its concepts.
 //!
 //! ## To individual handlers
 //!
 //! A middleware can be applied to a single handler like so:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     handler::{get, Handler},
+//!     Router,
+//! };
 //! use tower::limit::ConcurrencyLimitLayer;
 //!
-//! let app = route(
-//!     "/",
-//!     get(handler.layer(ConcurrencyLimitLayer::new(100))),
-//! );
+//! let app = Router::new()
+//!     .route(
+//!         "/",
+//!         get(handler.layer(ConcurrencyLimitLayer::new(100))),
+//!     );
 //!
 //! async fn handler() {}
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
@@ -428,10 +641,14 @@
 //! Middleware can also be applied to a group of routes like so:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     handler::{get, post},
+//!     Router,
+//! };
 //! use tower::limit::ConcurrencyLimitLayer;
 //!
-//! let app = route("/", get(get_slash))
+//! let app = Router::new()
+//!     .route("/", get(get_slash))
 //!     .route("/foo", post(post_foo))
 //!     .layer(ConcurrencyLimitLayer::new(100));
 //!
@@ -439,7 +656,7 @@
 //!
 //! async fn post_foo() {}
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
@@ -453,46 +670,50 @@
 //! However when applying middleware, or embedding other tower services, errors
 //! might happen. For example [`Timeout`] will return an error if the timeout
 //! elapses. By default these errors will be propagated all the way up to hyper
-//! where the connection will be closed. If that isn't desireable you can call
+//! where the connection will be closed. If that isn't desirable you can call
 //! [`handle_error`](handler::Layered::handle_error) to handle errors from
 //! adding a middleware to a handler:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     handler::{get, Handler},
+//!     Router,
+//! };
 //! use tower::{
 //!     BoxError, timeout::{TimeoutLayer, error::Elapsed},
 //! };
 //! use std::{borrow::Cow, time::Duration, convert::Infallible};
 //! use http::StatusCode;
 //!
-//! let app = route(
-//!     "/",
-//!     get(handle
-//!         .layer(TimeoutLayer::new(Duration::from_secs(30)))
-//!         // `Timeout` uses `BoxError` as the error type
-//!         .handle_error(|error: BoxError| {
-//!             // Check if the actual error type is `Elapsed` which
-//!             // `Timeout` returns
-//!             if error.is::<Elapsed>() {
-//!                 return Ok::<_, Infallible>((
-//!                     StatusCode::REQUEST_TIMEOUT,
-//!                     "Request took too long".into(),
-//!                 ));
-//!             }
+//! let app = Router::new()
+//!     .route(
+//!         "/",
+//!         get(handle
+//!             .layer(TimeoutLayer::new(Duration::from_secs(30)))
+//!             // `Timeout` uses `BoxError` as the error type
+//!             .handle_error(|error: BoxError| {
+//!                 // Check if the actual error type is `Elapsed` which
+//!                 // `Timeout` returns
+//!                 if error.is::<Elapsed>() {
+//!                     return Ok::<_, Infallible>((
+//!                         StatusCode::REQUEST_TIMEOUT,
+//!                         "Request took too long".into(),
+//!                     ));
+//!                 }
 //!
-//!             // If we encounter some error we don't handle return a generic
-//!             // error
-//!             return Ok::<_, Infallible>((
-//!                 StatusCode::INTERNAL_SERVER_ERROR,
-//!                 // `Cow` lets us return either `&str` or `String`
-//!                 Cow::from(format!("Unhandled internal error: {}", error)),
-//!             ));
-//!         })),
-//! );
+//!                 // If we encounter some error we don't handle return a generic
+//!                 // error
+//!                 return Ok::<_, Infallible>((
+//!                     StatusCode::INTERNAL_SERVER_ERROR,
+//!                     // `Cow` lets us return either `&str` or `String`
+//!                     Cow::from(format!("Unhandled internal error: {}", error)),
+//!                 ));
+//!             })),
+//!     );
 //!
 //! async fn handle() {}
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
@@ -500,14 +721,19 @@
 //! return `Result<T, E>` where `T` implements
 //! [`IntoResponse`](response::IntoResponse).
 //!
-//! See [`routing::Layered::handle_error`] fo more details.
+//! See [`routing::Router::handle_error`] for more details.
 //!
 //! ## Applying multiple middleware
 //!
 //! [`tower::ServiceBuilder`] can be used to combine multiple middleware:
 //!
 //! ```rust,no_run
-//! use axum::prelude::*;
+//! use axum::{
+//!     body::Body,
+//!     handler::get,
+//!     http::Request,
+//!     Router,
+//! };
 //! use tower::ServiceBuilder;
 //! use tower_http::compression::CompressionLayer;
 //! use std::{borrow::Cow, time::Duration};
@@ -523,10 +749,121 @@
 //!     .layer(CompressionLayer::new())
 //!     .into_inner();
 //!
-//! let app = route("/", get(|_: Request<Body>| async { /* ... */ }))
+//! let app = Router::new()
+//!     .route("/", get(|_: Request<Body>| async { /* ... */ }))
 //!     .layer(middleware_stack);
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # };
+//! ```
+//!
+//! ## Commonly used middleware
+//!
+//! [`tower::util`] and [`tower_http`] have a large collection of middleware that are compatible
+//! with axum. Some commonly used are:
+//!
+//! ```rust,no_run
+//! use axum::{
+//!     body::{Body, BoxBody},
+//!     handler::get,
+//!     http::{Request, Response},
+//!     Router,
+//! };
+//! use tower::{
+//!     filter::AsyncFilterLayer,
+//!     util::AndThenLayer,
+//!     ServiceBuilder,
+//! };
+//! use std::convert::Infallible;
+//! use tower_http::trace::TraceLayer;
+//!
+//! let middleware_stack = ServiceBuilder::new()
+//!     // `TraceLayer` adds high level tracing and logging
+//!     .layer(TraceLayer::new_for_http())
+//!     // `AsyncFilterLayer` lets you asynchronously transform the request
+//!     .layer(AsyncFilterLayer::new(map_request))
+//!     // `AndThenLayer` lets you asynchronously transform the response
+//!     .layer(AndThenLayer::new(map_response))
+//!     .into_inner();
+//!
+//! async fn map_request(req: Request<Body>) -> Result<Request<Body>, Infallible> {
+//!     Ok(req)
+//! }
+//!
+//! async fn map_response(res: Response<BoxBody>) -> Result<Response<BoxBody>, Infallible> {
+//!     Ok(res)
+//! }
+//!
+//! let app = Router::new()
+//!     .route("/", get(|| async { /* ... */ }))
+//!     .layer(middleware_stack);
+//! # async {
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # };
+//! ```
+//!
+//! Additionally axum provides [`extract::extractor_middleware()`] for converting any extractor into
+//! a middleware. Among other things, this can be useful for doing authorization. See
+//! [`extract::extractor_middleware()`] for more details.
+//!
+//! ## Writing your own middleware
+//!
+//! You can also write you own middleware by implementing [`tower::Service`]:
+//!
+//! ```
+//! use axum::{
+//!     body::{Body, BoxBody},
+//!     handler::get,
+//!     http::{Request, Response},
+//!     Router,
+//! };
+//! use futures::future::BoxFuture;
+//! use tower::{Service, layer::layer_fn};
+//! use std::task::{Context, Poll};
+//!
+//! #[derive(Clone)]
+//! struct MyMiddleware<S> {
+//!     inner: S,
+//! }
+//!
+//! impl<S, ReqBody, ResBody> Service<Request<ReqBody>> for MyMiddleware<S>
+//! where
+//!     S: Service<Request<ReqBody>, Response = Response<ResBody>> + Clone + Send + 'static,
+//!     S::Future: Send + 'static,
+//!     ReqBody: Send + 'static,
+//!     ResBody: Send + 'static,
+//! {
+//!     type Response = S::Response;
+//!     type Error = S::Error;
+//!     type Future = BoxFuture<'static, Result<Self::Response, Self::Error>>;
+//!
+//!     fn poll_ready(&mut self, cx: &mut Context<'_>) -> Poll<Result<(), Self::Error>> {
+//!         self.inner.poll_ready(cx)
+//!     }
+//!
+//!     fn call(&mut self, mut req: Request<ReqBody>) -> Self::Future {
+//!         println!("`MyMiddleware` called!");
+//!
+//!         // best practice is to clone the inner service like this
+//!         // see https://github.com/tower-rs/tower/issues/547 for details
+//!         let clone = self.inner.clone();
+//!         let mut inner = std::mem::replace(&mut self.inner, clone);
+//!
+//!         Box::pin(async move {
+//!             let res: Response<ResBody> = inner.call(req).await?;
+//!
+//!             println!("`MyMiddleware` received the response");
+//!
+//!             Ok(res)
+//!         })
+//!     }
+//! }
+//!
+//! let app = Router::new()
+//!     .route("/", get(|| async { /* ... */ }))
+//!     .layer(layer_fn(|inner| MyMiddleware { inner }));
+//! # async {
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
@@ -538,7 +875,12 @@
 //! and the [`extract::Extension`] extractor:
 //!
 //! ```rust,no_run
-//! use axum::{AddExtensionLayer, prelude::*};
+//! use axum::{
+//!     AddExtensionLayer,
+//!     extract,
+//!     handler::get,
+//!     Router,
+//! };
 //! use std::sync::Arc;
 //!
 //! struct State {
@@ -547,7 +889,9 @@
 //!
 //! let shared_state = Arc::new(State { /* ... */ });
 //!
-//! let app = route("/", get(handler)).layer(AddExtensionLayer::new(shared_state));
+//! let app = Router::new()
+//!     .route("/", get(handler))
+//!     .layer(AddExtensionLayer::new(shared_state));
 //!
 //! async fn handler(
 //!     state: extract::Extension<Arc<State>>,
@@ -557,99 +901,35 @@
 //!     // ...
 //! }
 //! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
-//! # };
-//! ```
-//!
-//! # Routing to any [`Service`]
-//!
-//! axum also supports routing to general [`Service`]s:
-//!
-//! ```rust,no_run
-//! use axum::{service, prelude::*};
-//! use tower_http::services::ServeFile;
-//! use http::Response;
-//! use std::convert::Infallible;
-//! use tower::{service_fn, BoxError};
-//!
-//! let app = route(
-//!     // Any request to `/` goes to a service
-//!     "/",
-//!     // Services who's response body is not `axum::body::BoxBody`
-//!     // can be wrapped in `axum::service::any` (or one of the other routing filters)
-//!     // to have the response body mapped
-//!     service::any(service_fn(|_: Request<Body>| async {
-//!         let res = Response::new(Body::from("Hi from `GET /`"));
-//!         Ok(res)
-//!     }))
-//! ).route(
-//!     "/foo",
-//!     // This service's response body is `axum::body::BoxBody` so
-//!     // it can be routed to directly.
-//!     service_fn(|req: Request<Body>| async move {
-//!         let body = Body::from(format!("Hi from `{} /foo`", req.method()));
-//!         let body = axum::body::box_body(body);
-//!         let res = Response::new(body);
-//!         Ok(res)
-//!     })
-//! ).route(
-//!     // GET `/static/Cargo.toml` goes to a service from tower-http
-//!     "/static/Cargo.toml",
-//!     service::get(ServeFile::new("Cargo.toml"))
-//! );
-//! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
-//! # };
-//! ```
-//!
-//! Routing to arbitrary services in this way has complications for backpressure
-//! ([`Service::poll_ready`]). See the [`service`] module for more details.
-//!
-//! # Nesting applications
-//!
-//! Applications can be nested by calling [`nest`](routing::nest):
-//!
-//! ```rust,no_run
-//! use axum::{prelude::*, routing::BoxRoute, body::{Body, BoxBody}};
-//! use tower_http::services::ServeFile;
-//! use http::Response;
-//!
-//! fn api_routes() -> BoxRoute<Body> {
-//!     route("/users", get(|_: Request<Body>| async { /* ... */ })).boxed()
-//! }
-//!
-//! let app = route("/", get(|_: Request<Body>| async { /* ... */ }))
-//!     .nest("/api", api_routes());
-//! # async {
-//! # hyper::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
+//! # axum::Server::bind(&"".parse().unwrap()).serve(app.into_make_service()).await.unwrap();
 //! # };
 //! ```
 //!
 //! # Required dependencies
 //!
-//! To use Axum there are a few dependencies you have pull in as well:
+//! To use axum there are a few dependencies you have pull in as well:
 //!
 //! ```toml
 //! [dependencies]
 //! axum = "<latest-version>"
-//!
-//! # For http types like `StatusCode`, `Method`, and `Response`
-//! http = "<latest-version>"
-//!
-//! # "full" isn't strictly necessary for tokio and hyper but its the
-//! # easiest way to get started.
 //! hyper = { version = "<latest-version>", features = ["full"] }
 //! tokio = { version = "<latest-version>", features = ["full"] }
-//!
-//! # Not strictly necessary but helpful for testing. There is a
-//! # testing example in the repo that shows how to test axum apps.
 //! tower = "<latest-version>"
 //! ```
 //!
+//! The `"full"` feature for hyper and tokio isn't strictly necessary but its
+//! the easiest way to get started.
+//!
+//! Note that [`axum::Server`] is re-exported by axum so if thats all you need
+//! then you don't have to explicitly depend on hyper.
+//!
+//! Tower isn't strictly necessary either but helpful for testing. See the
+//! testing example in the repo to learn more about testing axum apps.
+//!
 //! # Examples
 //!
 //! The axum repo contains [a number of examples][examples] that show how to put all the
-//! pieces togehter.
+//! pieces together.
 //!
 //! # Feature flags
 //!
@@ -659,8 +939,9 @@
 //! The following optional features are available:
 //!
 //! - `ws`: Enables WebSockets support.
-//! - `headers`: Enables extracing typed headers via [`extract::TypedHeader`].
+//! - `headers`: Enables extracting typed headers via [`extract::TypedHeader`].
 //! - `multipart`: Enables parsing `multipart/form-data` requests with [`extract::Multipart`].
+//! - `tower-log`: Enables `tower`'s `log` feature. Enabled by default.
 //!
 //! [`tower`]: https://crates.io/crates/tower
 //! [`tower-http`]: https://crates.io/crates/tower-http
@@ -671,8 +952,15 @@
 //! [`IntoResponse`]: crate::response::IntoResponse
 //! [`Timeout`]: tower::timeout::Timeout
 //! [examples]: https://github.com/tokio-rs/axum/tree/main/examples
+//! [`Router::or`]: crate::routing::Router::or
+//! [`axum::Server`]: hyper::server::Server
+//! [`OriginalUri`]: crate::extract::OriginalUri
+//! [`Service`]: tower::Service
+//! [`Service::poll_ready`]: tower::Service::poll_ready
+//! [`tower::Service`]: tower::Service
+//! [`handle_error`]: routing::Router::handle_error
+//! [tower-guides]: https://github.com/tower-rs/tower/tree/master/guides
 
-#![doc(html_root_url = "https://docs.rs/axum/0.1.2")]
 #![warn(
     clippy::all,
     clippy::dbg_macro,
@@ -707,26 +995,18 @@
     missing_debug_implementations,
     missing_docs
 )]
-#![deny(unreachable_pub, broken_intra_doc_links, private_in_public)]
-#![allow(
-    elided_lifetimes_in_paths,
-    // TODO: Remove this once the MSRV bumps to 1.42.0 or above.
-    clippy::match_like_matches_macro,
-    clippy::type_complexity
-)]
+#![deny(unreachable_pub, private_in_public)]
+#![allow(elided_lifetimes_in_paths, clippy::type_complexity)]
 #![forbid(unsafe_code)]
 #![cfg_attr(docsrs, feature(doc_cfg))]
 #![cfg_attr(test, allow(clippy::float_cmp))]
 
-use self::body::Body;
-use http::Request;
-use routing::{EmptyRouter, Route};
-use tower::Service;
-
 #[macro_use]
 pub(crate) mod macros;
 
 mod buffer;
+mod error;
+mod json;
 mod util;
 
 pub mod body;
@@ -735,75 +1015,21 @@ pub mod handler;
 pub mod response;
 pub mod routing;
 pub mod service;
-pub mod sse;
-
-#[cfg(feature = "ws")]
-#[cfg_attr(docsrs, doc(cfg(feature = "ws")))]
-pub mod ws;
 
 #[cfg(test)]
 mod tests;
 
+#[doc(no_inline)]
 pub use async_trait::async_trait;
+#[doc(no_inline)]
+pub use http;
+#[doc(no_inline)]
+pub use hyper::Server;
+#[doc(no_inline)]
 pub use tower_http::add_extension::{AddExtension, AddExtensionLayer};
 
-pub mod prelude {
-    //! Re-exports of important traits, types, and functions used with axum. Meant to be glob
-    //! imported.
+#[doc(inline)]
+pub use self::{error::Error, json::Json, routing::Router};
 
-    pub use crate::body::Body;
-    pub use crate::extract;
-    pub use crate::handler::{
-        any, connect, delete, get, head, options, patch, post, put, trace, Handler,
-    };
-    pub use crate::response;
-    pub use crate::route;
-    pub use crate::routing::RoutingDsl;
-    pub use http::Request;
-}
-
-/// Create a route.
-///
-/// `description` is a string of path segments separated by `/`. Each segment
-/// can be either concrete or a capture:
-///
-/// - `/foo/bar/baz` will only match requests where the path is `/foo/bar/bar`.
-/// - `/:foo` will match any route with exactly one segment _and_ it will
-/// capture the first segment and store it at the key `foo`.
-///
-/// `service` is the [`Service`] that should receive the request if the path
-/// matches `description`.
-///
-/// # Examples
-///
-/// ```rust
-/// use axum::prelude::*;
-/// # use std::convert::Infallible;
-/// # use http::Response;
-/// # let service = tower::service_fn(|_: Request<Body>| async {
-/// #     Ok::<Response<Body>, Infallible>(Response::new(Body::empty()))
-/// # });
-///
-/// route("/", service);
-/// route("/users", service);
-/// route("/users/:id", service);
-/// route("/api/:version/users/:id/action", service);
-/// ```
-///
-/// # Panics
-///
-/// Panics if `description` doesn't start with `/`.
-pub fn route<S, B>(description: &str, service: S) -> Route<S, EmptyRouter<S::Error>>
-where
-    S: Service<Request<B>> + Clone,
-{
-    use routing::RoutingDsl;
-
-    routing::EmptyRouter::not_found().route(description, service)
-}
-
-mod sealed {
-    #![allow(unreachable_pub, missing_docs)]
-
-    pub trait Sealed {}
-}
+/// Alias for a type-erased error type.
+pub type BoxError = Box<dyn std::error::Error + Send + Sync>;
